﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Net;
using System.Xml;
using System.Web.Script.Serialization;
using System.Collections;
using System.Windows;

namespace WSI_sample
{
    /*
     * This class implements the logic. It uses the singleton pattern.
     */
    class Session
    {
        private static Session instance;        // instance of the session
        private Data data;                      // reference to the data instance
        private Call currentCall;               // Reference to the last call object, to edit it later. If not null, there is a current ringing or established call.
        private Communication communication;    // reference to the communication instance

        private Session()
        {
            data = Data.getData();
            communication = new Communication();
            currentCall = null;
        }

        public static Session getSession()
        {
            if (instance == null)
                instance = new Session();
            return instance;
        }

        /*
         * This method is called when the login button is clicked.
         */
        public void login()
        {
            // set the parameters
            Dictionary<String, String> parameters = new Dictionary<String, String>();
            parameters.Add("cmd", "Login");
            parameters.Add("gsUser", data.UserId);
            parameters.Add("gsPass", data.Password);

            data.TryLogin = true;

            communication.get(parameters, new AsyncCallback(handleLogin));
        }

        /*
         * This method is called, when the login request returns.
         */
        private void handleLogin(IAsyncResult result)
        {
            HttpWebResponse response;
            try
            {
                response = (result.AsyncState as HttpWebRequest).EndGetResponse(result) as HttpWebResponse;
            }
            catch (WebException e)
            {
                // there was an exception while making the request or the request was canceled
                data.TryLogin = false;
                data.LoggedIn = false;
                handleFailure(e);
                return;
            }

            // check the HTTP response code
            HttpStatusCode statusCode = response.StatusCode;
            if (statusCode == HttpStatusCode.OK)
            {
                StreamReader sr = new StreamReader(response.GetResponseStream());
                String xmlResponse = sr.ReadToEnd();
                response.Close();

                // add the response message to the list
                Message msg = new Message(Message.Command.Login, response.ResponseUri.ToString(), Tools.formatXml(xmlResponse, 3));
                data.addMessage(msg);

                // check, if the response has the expected content type
                String contentType = response.ContentType;
                if (contentType.Contains("text/xml"))
                {
                    XmlDocument xmlDoc = new XmlDocument();
                    xmlDoc.LoadXml(xmlResponse);

                    String sessionID = xmlDoc.SelectSingleNode("LOGIN/ID").InnerText;
                    // if the returned ID is not 0, the login was successful and you can call eventStart
                    if (sessionID != "0")
                    {
                        // store the gsSession ID in data
                        data.LoggedIn = true;
                        data.SessionId = sessionID;
                        data.TryLogin = false;

                        // call eventStart
                        eventStart();
                    }
                    else
                    {
                        data.LoggedIn = false;
                        data.TryLogin = false;

                        // if the ID is 0, the response contains an ERROR element
                        String error = xmlDoc.SelectSingleNode("LOGIN/ERROR").InnerText;
                        switch (error)
                        {
                            case "LOGIN_FAILED":
                                // Password wrong or user unknown, user not allowed to login.
                                break;
                            case "SERVICE_STARTING":
                                // The server is starting up, login should be retried after 30 seconds.
                                break;
                            case "LOGIN_PWWEAK":
                                // The user has to configure a new password before the Web Services Interface can be utilized.
                                break;
                        }
                    }
                }
                else
                {
                    handleContentTypeError(response);
                }

            }
            else
            {
                handleStatusCodeError(response);
            }
        }

        /*
         * This method is called, when the logout button is clicked.
         */
        public void logout()
        {
            // set the parameters, remember to hand over the sessionID
            Dictionary<String, String> parameters = new Dictionary<String, String>();
            parameters.Add("cmd", "Logout");
            parameters.Add("gsSession", data.SessionId);

            communication.get(parameters, new AsyncCallback(handleLogout));
        }

        /*
         * This method is called, when the logout request returns
         */
        private void handleLogout(IAsyncResult result)
        {
            HttpWebResponse response;
            try
            {
                response = (result.AsyncState as HttpWebRequest).EndGetResponse(result) as HttpWebResponse;
            }
            catch (WebException e)
            {
                // there was an exception while making the request or the request was canceled
                handleFailure(e);
                return;
            }

            // check the HTTP response code
            HttpStatusCode statusCode = response.StatusCode;
            if (statusCode == HttpStatusCode.OK)
            {
                StreamReader sr = new StreamReader(response.GetResponseStream());
                String xmlResponse = sr.ReadToEnd();
                response.Close();

                // add the response to the list
                Message msg = new Message(Message.Command.Logout, response.ResponseUri.ToString(), Tools.formatXml(xmlResponse, 3));
                data.addMessage(msg);

                // check, if the response has the expected content type
                String contentType = response.ContentType;
                if (contentType.Contains("text/xml"))
                {
                    XmlDocument xmlDoc = new XmlDocument();
                    xmlDoc.LoadXml(xmlResponse);

                    // check, if the XML document starts with the expected element
                    if (xmlDoc.SelectSingleNode("/*").Name == "LOGOUT")
                    {
                        // logout was successful
                        data.LoggedIn = false;
                    }
                    else
                    {
                        handleXmlError(response.ResponseUri.ToString(), xmlResponse);
                    }
                }
                else
                {
                    handleContentTypeError(response);
                }

            }
            else
            {
                handleStatusCodeError(response);
            }
        }

        /*
         * This method is called, when the login was successful.
         */
        public void eventStart()
        {
            // Note: If your system is older than V1 R3 use 'MonitorStart' instead.

            // set the parameters, remember to hand over the sessionID
            Dictionary<String, String> parameters = new Dictionary<String, String>();
            parameters.Add("cmd", "EventStart");
            parameters.Add("gsSession", data.SessionId);
            parameters.Add("device", data.UserId);
            parameters.Add("filter", "DEVSTATE");

            communication.get(parameters, new AsyncCallback(handleEventStart));
        }

        /*
         * This method is called, when the EventStart request returns.
         */
        private void handleEventStart(IAsyncResult result)
        {
            HttpWebResponse response;
            try
            {
                response = (result.AsyncState as HttpWebRequest).EndGetResponse(result) as HttpWebResponse;
            }
            catch (WebException e)
            {
                // there was an exception while making the request or the request was canceled
                handleFailure(e);
                return;
            }

            // check the HTTP response code
            HttpStatusCode statusCode = response.StatusCode;
            if (statusCode == HttpStatusCode.OK)
            {
                StreamReader sr = new StreamReader(response.GetResponseStream());
                String xmlResponse = sr.ReadToEnd();
                response.Close();

                //add the response to the list
                Message msg = new Message(Message.Command.EventStart, response.ResponseUri.ToString(), Tools.formatXml(xmlResponse, 3));
                data.addMessage(msg);

                // check, if the response has the expected content type
                String contentType = response.ContentType;
                if (contentType.Contains("text/xml"))
                {
                    XmlDocument xmlDoc = new XmlDocument();
                    xmlDoc.LoadXml(xmlResponse);

                    // check, if the XML document starts with the expected element
                    if (xmlDoc.SelectSingleNode("/*").Name == "DONE")
                    {
                        // EventStart was successful, start receiving events
                        getEvents();
                    }
                    else
                    {
                        handleXmlError(response.ResponseUri.ToString(), xmlResponse);
                    }
                }
                else
                {
                    handleContentTypeError(response);
                }

            }
            else
            {
                handleStatusCodeError(response);
            }
        }

        /*
         * This method is called, when the EventStart request returned with no errors.
         */
        public void getEvents()
        {
            // set the parameters, remember to hand over the sessionID
            Dictionary<String, String> parameters = new Dictionary<String, String>();
            parameters.Add("cmd", "GetEvents");
            parameters.Add("gsSession", data.SessionId);
            parameters.Add("deviceObject", data.UserId);

            communication.get(parameters, new AsyncCallback(handleGetEvents));
        }

        /*
         * This method is called, when the GetEvents request returns.
         */
        private void handleGetEvents(IAsyncResult result)
        {
            HttpWebResponse response;
            try
            {
                response = (result.AsyncState as HttpWebRequest).EndGetResponse(result) as HttpWebResponse;
            }
            catch (WebException e)
            {
                // there was an exception while making the request or the request was canceled
                handleFailure(e);
                return;
            }

            // check the HTTP response code
            HttpStatusCode statusCode = response.StatusCode;
            if (statusCode == HttpStatusCode.OK)
            {
                StreamReader sr = new StreamReader(response.GetResponseStream());
                String plainResponse = sr.ReadToEnd();
                response.Close();
                Message msg = null;

                // extract contentType of the response
                String contentType = response.ContentType;
                contentType = contentType.Split(';')[0]; // contentType looks like 'text/html; charset=utf-8'

                // The response could have one of the following types:
                //      application/json: the request returned, because an event occurred
                //      text/plain: no event occured and the request returned after 30 seconds
                //      text/xml: an error occured
                switch (contentType)
                {
                    case "application/json":
                        // the request returned, because an event occured

                        msg = new Message(Message.Command.GetEvents, response.ResponseUri.ToString(), Tools.formatJson(plainResponse));
                        data.addMessage(msg);

                        String delta = "";
                        try
                        {
                            // extract the delta from the JSON response
                            JavaScriptSerializer ser = new JavaScriptSerializer();
                            Dictionary<String, object> dict = ser.Deserialize<Dictionary<String, object>>(plainResponse);

                            ArrayList events = (ArrayList)dict["events"];
                            Dictionary<String, object> firstEvent = (Dictionary<String, object>)events[0];
                            Dictionary<String, object> jsonEvent = (Dictionary<String, object>)firstEvent["jsonEvent"];
                            Dictionary<String, object> DevStateChange = (Dictionary<String, object>)jsonEvent["DevStateChange"];
                            delta = (String)DevStateChange["delta"];
                        }
                        catch (Exception e) { Console.WriteLine(e.StackTrace); }

                        // if the delta contains 'call#, get information about the device state
                        if (delta.Contains("call"))
                            getDevState();

                        // query GetEvents again to receive further events
                        getEvents();
                        break;
                    case "text/plain":
                        // the request returned after 30 seconds with no events and should be queried again, to receive further events

                        msg = new Message(Message.Command.GetEvents, response.ResponseUri.ToString(), plainResponse);
                        data.addMessage(msg);

                        getEvents();
                        break;
                    case "text/xml":
                        // the request returned because an error occurred

                        msg = new Message(Message.Command.GetEvents, response.ResponseUri.ToString(), Tools.formatXml(plainResponse, 3));
                        data.addMessage(msg);

                        break;
                }
            }
            else
            {
                handleStatusCodeError(response);
            }
        }

        /*
         * This method is called, when an events was received.
         */
        public void getDevState()
        {
            // set the parameters, remember to hand over the sessionID
            Dictionary<String, String> parameters = new Dictionary<String, String>();
            parameters.Add("cmd", "GetDeviceState");
            parameters.Add("gsSession", data.SessionId);
            parameters.Add("device", data.UserId);

            communication.get(parameters, new AsyncCallback(handleGetDevState));
        }

        /*
         * This method is called, when the GetDeviceState request returns.
         */
        private void handleGetDevState(IAsyncResult result)
        {
            HttpWebResponse response;
            try
            {
                response = (result.AsyncState as HttpWebRequest).EndGetResponse(result) as HttpWebResponse;
            }
            catch (WebException e)
            {
                // there was an exception while making the request or the request was canceled
                handleFailure(e);
                return;
            }

            // check the HTTP response code
            HttpStatusCode statusCode = response.StatusCode;
            if (statusCode == HttpStatusCode.OK)
            {
                StreamReader sr = new StreamReader(response.GetResponseStream());
                String xmlResponse = sr.ReadToEnd();
                response.Close();

                // add the response to the list
                Message msg = new Message(Message.Command.GetDeviceState, response.ResponseUri.ToString(), Tools.formatXml(xmlResponse, 3));
                data.addMessage(msg);

                // check, if the response has the expected content type
                String contentType = response.ContentType;
                if (contentType.Contains("text/xml"))
                {
                    XmlDocument xmlDoc = new XmlDocument();
                    xmlDoc.LoadXml(xmlResponse);

                    // check, if the XML document starts with the expected element
                    if (xmlDoc.SelectSingleNode("/*").Name == "DevState")
                    {
                        // If it was a call event, the callList will have an entry about the call.
                        XmlNode callList = xmlDoc.SelectSingleNode("DevState/CallList");
                        if (callList != null)
                        {
                            XmlNode call = callList.SelectSingleNode("Call");
                            if (call != null)
                            {
                                String state = call.SelectSingleNode("State").InnerText;
                                DateTime date = DateTime.Now;

                                switch (state)
                                {
                                    case "Ringing":
                                        String caller = call.SelectSingleNode("Calling").Attributes["name"].Value; // get the name of the caller
                                        if (caller == "" || caller == null)
                                            caller = call.SelectSingleNode("Calling").InnerText;    // use phone number if no name is present

                                        String called = call.SelectSingleNode("Called").Attributes["name"].Value;   // get the name of the called person
                                        if (called == "" || caller == null)
                                            called = call.SelectSingleNode("Called").InnerText;     // use phone number if no name is present

                                        currentCall = new Call(caller, called, date.Ticks);

                                        // check if it was an incoming call
                                        XmlNodeList flags = call.SelectSingleNode("Flags").SelectNodes("F");
                                        bool isIncoming = false;
                                        foreach (XmlNode flag in flags)
                                        {
                                            if (flag.InnerText == "Incoming")
                                                isIncoming = true;
                                        }

                                        currentCall.Incoming = isIncoming;

                                        break;
                                    case "Established":
                                        currentCall.From = date.Ticks;
                                        break;
                                }
                            }
                            else
                            {
                                // This block is executed, when the response does not have a call entry.
                                // If 'currentCall' is not null, this indicates, that the call was ended.
                                XmlNode ops = xmlDoc.SelectSingleNode("DevState/CallList").SelectSingleNode("Ops");
                                if (ops != null)
                                {
                                    if (ops.InnerText == "makecall")
                                    {
                                        if (currentCall != null)
                                        {
                                            DateTime date = DateTime.Now;
                                            currentCall.To = date.Ticks;
                                            data.addCall(currentCall);
                                            currentCall = null;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else
                    {
                        handleXmlError(response.ResponseUri.ToString(), xmlResponse);
                    }
                }
                else
                {
                    handleContentTypeError(response);
                }

            }
            else
            {
                handleStatusCodeError(response);
            }
        }


        /*
         * This method is called, when a response has a HTTP StatusCode different to 200.
         */
        public void handleStatusCodeError(HttpWebResponse response)
        {
            Console.WriteLine("url: " + response.ResponseUri);
            Console.WriteLine("statusCode: " + response.StatusCode.ToString());
            response.Close();
        }

        /*
         * This method is called, when a response has an unexpected content type.
         */
        public void handleContentTypeError(HttpWebResponse response)
        {
            Console.WriteLine("url: " + response.ResponseUri);
            Console.WriteLine("contentType: " + response.ContentType);
            response.Close();
        }

        /*
         * This method is called, when a xml response starts with an unexpected element. 
         */
        public void handleXmlError(String url, String response)
        {
            Console.WriteLine("url: " + url);
            Console.WriteLine("contentType: " + response);
        }

        /*
         * This method is called, when an exception occurs while making a HTTP request.
         */
        public void handleFailure(Exception e)
        {
            Console.Write(e.StackTrace);
        }
    }
}
